@@ -1,10 +1,6 @@
 use crate::*;
 use gdnative::core_types::ToVariant;
-use ggrs::{
-    Frame, GGRSEvent, GGRSRequest, GameState, GameStateCell, P2PSession, PlayerHandle, PlayerType,
-    SessionState,
-};
-use std::convert::TryInto;
+use ggrs::{Frame, GGRSEvent, P2PSession, PlayerHandle, PlayerType, SessionState};
 use std::option::*;
 
 /// A Godot implementation of [`P2PSession`]
@@ -35,32 +31,48 @@ impl GodotGGRSP2PSession {
     }
 
     /// Creates a [P2PSession],
-    /// call this when you want to start setting up a P2P Session takes the local port and total number of players as parameters
+    /// call this when you want to start setting up a P2P Session takes the local port, total number of players and max prediction frames as parameters.
+    /// # Notes
+    /// - Max prediction frames is the maximum number of frames GGRS will roll back. Every gamestate older than this is guaranteed to be correct if the players did not desync.
+    /// - This value used to default to `8 frames`, but this has been made adjustable with `GGRS 0.7.0`
     #[export]
-    pub fn create_session(&mut self, _owner: &Node, local_port: u16, num_players: u32) {
+    pub fn create_new_session(
+        &mut self,
+        _owner: &Node,
+        local_port: u16,
+        num_players: u32,
+        max_pred: usize,
+    ) {
         let input_size: usize = std::mem::size_of::<u32>();
-        match P2PSession::new(num_players, input_size, local_port) {
+        match P2PSession::new(num_players, input_size, max_pred, local_port) {
             Ok(s) => self.sess = Some(s),
             Err(e) => godot_error!("{}", e),
         }
     }
 
-    /// Adds a local player to the a session and return the handle
+    /// Deprecated method to create a [P2PSession]. Use [Self::create_new_session()] instead.
+    #[deprecated(since = "0.5.0", note = "please use `create_new_session()` instead")]
+    #[export]
+    pub fn create_session(&mut self, _owner: &Node, local_port: u16, num_players: u32) {
+        self.create_new_session(_owner, local_port, num_players, 8)
+    }
+
+    /// Adds a local player to the session and return the handle.
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn add_local_player(&mut self, _owner: &Node) -> PlayerHandle {
         self.add_player(PlayerType::Local)
     }
 
-    /// Adds a remote player to the a session and returns the handle
+    /// Adds a remote player to the session and returns the handle.
     /// # Example
     /// The following example shows how to format an address string, starting with the IP and ending with the port.
     /// ```
     /// p2p.add_remote_player("127.0.0.1:7070")
     /// ```
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     /// - Will panic if the address string could not be converted to an [std::net::SocketAddr]
     #[export]
     pub fn add_remote_player(&mut self, _owner: &Node, address: String) -> PlayerHandle {
@@ -68,9 +80,9 @@ impl GodotGGRSP2PSession {
         self.add_player(PlayerType::Remote(remote_addr))
     }
 
-    /// Adds a spectator to the a session and returns the handle
+    /// Adds a spectator to the session and returns the handle
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     /// - Will panic if the address string could not be converted to an [std::net::SocketAddr]
     #[export]
     pub fn add_spectator(&mut self, _owner: &Node, address: String) -> PlayerHandle {
@@ -80,7 +92,7 @@ impl GodotGGRSP2PSession {
 
     /// Starts the [P2PSession]
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn start_session(&mut self, _owner: &Node) {
         match &mut self.sess {
@@ -127,32 +139,37 @@ impl GodotGGRSP2PSession {
     /// - [CALLBACK_FUNC_LOAD_GAME_STATE]
     /// - [CALLBACK_FUNC_SAVE_GAME_STATE]
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
-    /// - Will print an [ERR_MESSAGE_NO_CALLBACK_NODE] error if a callback node has not been set
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_CALLBACK_NODE] error if a callback node has not been set
     #[export]
     pub fn advance_frame(&mut self, _owner: &Node, local_player_handle: usize, local_input: u32) {
         //Convert local_input into a byte array
         let local_input_bytes = local_input.to_be_bytes();
         let local_input_array_slice: &[u8] = &local_input_bytes[..];
 
-        match &mut self.sess {
-            Some(s) => match s.advance_frame(local_player_handle, local_input_array_slice) {
-                Ok(requests) => {
-                    self.handle_requests(requests);
-                }
-                Err(e) => {
-                    godot_error!("{}", e);
+        match self.callback_node {
+            Some(callback_node) => match &mut self.sess {
+                Some(s) => match s.advance_frame(local_player_handle, local_input_array_slice) {
+                    Ok(requests) => {
+                        ggrs_request_handlers::handle_requests(&callback_node, requests);
+                    }
+                    Err(e) => {
+                        godot_error!("{}", e);
+                    }
+                },
+                None => {
+                    godot_error!("{}", ERR_MESSAGE_NO_SESSION_MADE);
                 }
             },
             None => {
-                godot_error!("{}", ERR_MESSAGE_NO_SESSION_MADE);
+                godot_error!("{}", ERR_MESSAGE_NO_CALLBACK_NODE);
             }
         }
     }
 
     /// Sets [P2PSession::set_fps()]
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn set_fps(&mut self, _owner: &Node, fps: u32) {
         match &mut self.sess {
@@ -172,7 +189,7 @@ impl GodotGGRSP2PSession {
 
     /// Calls [P2PSession::poll_remote_clients()]
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn poll_remote_clients(&mut self, _owner: &Node) {
         match &mut self.sess {
@@ -183,7 +200,7 @@ impl GodotGGRSP2PSession {
 
     /// Prints out network stats of specified handle
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn print_network_stats(&mut self, _owner: &Node, handle: PlayerHandle) {
         match &mut self.sess {
@@ -197,7 +214,7 @@ impl GodotGGRSP2PSession {
 
     /// Will return network stats of specified handle as a `tuple`, which will be converted to an `Array` inside godot.
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn get_network_stats(
         &mut self,
@@ -228,7 +245,7 @@ impl GodotGGRSP2PSession {
 
     /// Sets [P2PSession::set_frame_delay()] of specified handle.
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn set_frame_delay(
         &mut self,
@@ -247,7 +264,7 @@ impl GodotGGRSP2PSession {
 
     /// Sets [P2PSession::set_disconnect_timeout()] converting the u64 to secconds.
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn set_disconnect_timeout(&mut self, _owner: &Node, secs: u64) {
         match &mut self.sess {
@@ -258,7 +275,7 @@ impl GodotGGRSP2PSession {
 
     /// Sets [P2PSession::set_disconnect_notify_delay()] converting the u64 to secconds.
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn set_disconnect_notify_delay(&mut self, _owner: &Node, secs: u64) {
         match &mut self.sess {
@@ -269,7 +286,7 @@ impl GodotGGRSP2PSession {
 
     /// Sets [P2PSession::set_sparse_saving()].
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn set_sparse_saving(&mut self, _owner: &Node, sparse_saving: bool) {
         match &mut self.sess {
@@ -283,7 +300,7 @@ impl GodotGGRSP2PSession {
 
     /// Disconnects specified player handle.
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn disconnect_player(&mut self, _owner: &Node, player_handle: PlayerHandle) {
         match &mut self.sess {
@@ -319,7 +336,7 @@ impl GodotGGRSP2PSession {
     ///             var count = item[1][2]
     /// ```
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn get_events(&mut self, _owner: &Node) -> Vec<(&str, Variant)> {
         let mut result: Vec<(&str, Variant)> = Vec::new();
@@ -363,7 +380,7 @@ impl GodotGGRSP2PSession {
     /// Calls and returns [P2PSession::frames_ahead()].
     /// Will return a 0 if no session was made.
     /// # Errors
-    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    /// - Will print a [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
     #[export]
     pub fn get_frames_ahead(&mut self, _owner: &Node) -> i32 {
         match &mut self.sess {
@@ -375,85 +392,52 @@ impl GodotGGRSP2PSession {
         }
     }
 
+    /// Calls and returns [P2PSession::max_prediction()].
+    /// Will return a 0 if no session was made.
+    /// # Errors
+    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    #[export]
+    pub fn get_max_prediction(&mut self, _owner: &Node) -> usize {
+        match &mut self.sess {
+            Some(s) => s.max_prediction(),
+            None => {
+                godot_error!("{}", ERR_MESSAGE_NO_SESSION_MADE);
+                return 0;
+            }
+        }
+    }
+
+    /// Calls and returns [P2PSession::current_frame()].
+    /// Will return a 0 if no session was made.
+    /// # Errors
+    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    #[export]
+    pub fn get_current_frame(&mut self, _owner: &Node) -> Frame {
+        match &mut self.sess {
+            Some(s) => s.current_frame(),
+            None => {
+                godot_error!("{}", ERR_MESSAGE_NO_SESSION_MADE);
+                return 0;
+            }
+        }
+    }
+
+    /// Calls and returns [P2PSession::confirmed_frame()].
+    /// Will return a 0 if no session was made.
+    /// # Errors
+    /// - Will print an [ERR_MESSAGE_NO_SESSION_MADE] error if a session has not been made
+    #[export]
+    pub fn get_confirmed_frame(&mut self, _owner: &Node) -> Frame {
+        match &mut self.sess {
+            Some(s) => s.confirmed_frame(),
+            None => {
+                godot_error!("{}", ERR_MESSAGE_NO_SESSION_MADE);
+                return 0;
+            }
+        }
+    }
+
     //NON-EXPORTED FUNCTIONS
-    fn handle_requests(&mut self, requests: Vec<GGRSRequest>) {
-        for item in requests {
-            match item {
-                GGRSRequest::AdvanceFrame { inputs } => self.ggrs_request_advance_fame(inputs),
-                GGRSRequest::LoadGameState { cell } => self.ggrs_request_load_game_state(cell),
-                GGRSRequest::SaveGameState { cell, frame } => {
-                    self.ggrs_request_save_game_state(cell, frame);
-                }
-            }
-        }
-    }
-
-    fn ggrs_request_advance_fame(&self, inputs: Vec<ggrs::GameInput>) {
-        //Parse parameter inputs in a way that godot can handle then call the callback method
-        match self.callback_node {
-            Some(s) => {
-                let node = unsafe { s.assume_safe() };
-                let mut godot_array: Vec<Variant> = Vec::new();
-                for i in inputs {
-                    let result = (
-                        i.frame,
-                        i.size,
-                        u32::from_be_bytes(
-                            i.buffer[..i.size]
-                                .try_into()
-                                .expect("Slice size is too big or too small to convert into u32"),
-                        ),
-                    )
-                        .to_variant();
-                    godot_array.push(result);
-                }
-                unsafe { node.call(CALLBACK_FUNC_ADVANCE_FRAME, &[godot_array.to_variant()]) };
-            }
-            None => {
-                godot_error!("{}", ERR_MESSAGE_NO_CALLBACK_NODE);
-            }
-        }
-    }
-
-    fn ggrs_request_load_game_state(&self, cell: GameStateCell) {
-        //Unpack the cell and have over it's values to godot so it can handle it.
-        match self.callback_node {
-            Some(s) => {
-                let node = unsafe { s.assume_safe() };
-                let game_state = cell.load();
-                let frame = game_state.frame.to_variant();
-                let buffer =
-                    ByteArray::from_vec(game_state.buffer.unwrap_or_default()).to_variant();
-                let checksum = game_state.checksum.to_variant();
-                unsafe { node.call(CALLBACK_FUNC_LOAD_GAME_STATE, &[frame, buffer, checksum]) };
-            }
-            None => {
-                godot_error!("{}", ERR_MESSAGE_NO_CALLBACK_NODE);
-            }
-        }
-    }
-
-    fn ggrs_request_save_game_state(&mut self, cell: GameStateCell, frame: Frame) {
-        //Store current cell for later use
-        match self.callback_node {
-            Some(s) => {
-                let node = unsafe { s.assume_safe() };
-                let state: Variant =
-                    unsafe { node.call(CALLBACK_FUNC_SAVE_GAME_STATE, &[frame.to_variant()]) };
-                let state_bytes = ByteArray::from_variant(&state).unwrap_or_default();
-                let mut state_bytes_vec = Vec::new();
-                for i in 0..state_bytes.len() {
-                    state_bytes_vec.push(state_bytes.get(i));
-                }
-                let result = GameState::new(frame, Some(state_bytes_vec), None);
-                cell.save(result);
-            }
-            None => {
-                godot_error!("{}", ERR_MESSAGE_NO_CALLBACK_NODE);
-            }
-        }
-    }
-
     fn add_player(&mut self, player_type: PlayerType) -> PlayerHandle {
         match &mut self.sess {
             Some(s) => match s.add_player(player_type, self.next_handle) {
